# Copyright 2021 Erfan Abdi
# Copyright 2025 Bardia Moshiri
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import shutil
import os
import tools.config
from tools import helpers
from shutil import which

property_keys = [
    "ro.product.vendor.name",
    "ro.product.vendor.brand",
    "ro.product.vendor.model",
    "ro.product.vendor.device",
    "ro.product.vendor.manufacturer",
    "ro.product.vendor_dlkm.name",
    "ro.product.vendor_dlkm.brand",
    "ro.product.vendor_dlkm.model",
    "ro.product.vendor_dlkm.device",
    "ro.product.vendor_dlkm.manufacturer",
    "ro.product.manufacturer",
    "ro.product.name",
    "ro.product.brand",
    "ro.product.model",
    "ro.product.device",
    "ro.product.odm.name",
    "ro.product.odm.brand",
    "ro.product.odm.model",
    "ro.product.odm.device",
    "ro.product.odm.manufacturer",
    "ro.soc.model",
    "ro.soc.manufacturer",
    "ro.vendor.soc.model",
    "ro.vendor.soc.model.part_name",
    "ro.vendor.soc.model.external_name",

    # MediaTek specific props from here
    "ro.vendor.mediatek.platform",
    "ro.vendor.mediatek.version.branch",
    "ro.vendor.mediatek.version.release",
    "ro.mediatek.version.branch"
]

def remove_overlay(args):
    if os.path.isdir(tools.config.defaults["overlay_rw"]):
        shutil.rmtree(tools.config.defaults["overlay_rw"])
    if os.path.isdir(tools.config.defaults["overlay_work"]):
        shutil.rmtree(tools.config.defaults["overlay_work"])

def make_prop(args, cfg, full_props_path):
    if not os.path.isfile(args.work + "/andromeda_base.prop"):
        raise RuntimeError("andromeda_base.prop Not found")
    with open(args.work + "/andromeda_base.prop") as f:
        props = f.read().splitlines()
    if not props:
        raise RuntimeError("andromeda_base.prop is broken!!?")

    def add_prop(key, cfg_key):
        value = cfg[cfg_key]
        if value != "None":
            value = value.replace("/mnt/", "/mnt_extra/")
            props.append(key + "=" + value)

    add_prop("waydroid.host.user", "user_name")
    add_prop("waydroid.host.uid", "user_id")
    add_prop("waydroid.host.gid", "group_id")
    add_prop("waydroid.host_data_path", "andromeda_data")
    add_prop("waydroid.background_start", "background_start")
    props.append("waydroid.xdg_runtime_dir=" + tools.config.defaults["container_xdg_runtime_dir"])
    props.append("waydroid.pulse_runtime_path=" + tools.config.defaults["container_pulse_runtime_path"])
    props.append("waydroid.wayland_display=" + tools.config.defaults["container_wayland_display"])

    add_prop("andromeda.host.user", "user_name")
    add_prop("andromeda.host.uid", "user_id")
    add_prop("andromeda.host.gid", "group_id")
    add_prop("andromeda.host_data_path", "andromeda_data")
    add_prop("andromeda.background_start", "background_start")
    props.append("andromeda.xdg_runtime_dir=" + tools.config.defaults["container_xdg_runtime_dir"])
    props.append("andromeda.pulse_runtime_path=" + tools.config.defaults["container_pulse_runtime_path"])
    props.append("andromeda.wayland_display=" + tools.config.defaults["container_wayland_display"])

    for prop in property_keys:
        value = helpers.props.host_get(args, prop)
        if value:
            props.append(prop + "=" + value)

    if not os.path.exists("/usr/libexec/andromeda-sensors"):
        props.append("andromeda.stub_sensors_hal=1")

    if not os.path.exists("/usr/libexec/android-vibrator"):
        props.append("furios.stub_vibrator_hal=1")

    dpi = cfg["lcd_density"]
    if dpi != "0":
        props.append("ro.sf.lcd_density=" + dpi)

    width = cfg["width"]
    if width != "0":
        props.append("waydroid.display_width_override=" + width)
        props.append("andromeda.display_width_override=" + width)

    height = cfg["height"]
    if height != "0":
        props.append("waydroid.display_height_override=" + height)
        props.append("andromeda.display_width_override=" + width)

    final_props = open(full_props_path, "w")
    for prop in props:
        final_props.write(prop + "\n")
    final_props.close()
    os.chmod(full_props_path, 0o644)

def mount_rootfs(args, images_dir, session):
    cfg = tools.config.load(args)
    helpers.mount.mount(args, images_dir + "/system.img",
                        tools.config.defaults["rootfs"], umount=True)
    if cfg["andromeda"]["mount_overlays"] == "True":
        try:
            helpers.mount.mount_overlay(args, [tools.config.defaults["overlay"],
                                               tools.config.defaults["rootfs"]],
                                    tools.config.defaults["rootfs"],
                                    upper_dir=tools.config.defaults["overlay_rw"] + "/system",
                                    work_dir=tools.config.defaults["overlay_work"] + "/system")
        except RuntimeError:
            cfg["andromeda"]["mount_overlays"] = "False"
            tools.config.save(args, cfg)
            logging.warning("Mounting overlays failed. The feature has been disabled.")

    helpers.mount.mount(args, images_dir + "/vendor.img",
                           tools.config.defaults["rootfs"] + "/vendor")
    if cfg["andromeda"]["mount_overlays"] == "True":
        helpers.mount.mount_overlay(args, [tools.config.defaults["overlay"] + "/vendor",
                                           tools.config.defaults["rootfs"] + "/vendor"],
                                    tools.config.defaults["rootfs"] + "/vendor",
                                    upper_dir=tools.config.defaults["overlay_rw"] + "/vendor",
                                    work_dir=tools.config.defaults["overlay_work"] + "/vendor")

    for egl_path in ["/vendor/lib/egl", "/vendor/lib64/egl"]:
        if os.path.isdir(egl_path):
            helpers.mount.bind(
                args, egl_path, tools.config.defaults["rootfs"] + egl_path)
    if helpers.mount.ismount("/odm"):
        helpers.mount.bind(
            args, "/odm", tools.config.defaults["rootfs"] + "/odm_extra")
    else:
        if os.path.isdir("/vendor/odm"):
            helpers.mount.bind(
                args, "/vendor/odm", tools.config.defaults["rootfs"] + "/odm_extra")

    vendor_andromeda_prop = tools.config.defaults["rootfs"] + "/vendor/andromeda.prop"
    vendor_waydroid_prop = tools.config.defaults["rootfs"] + "/vendor/waydroid.prop"
    make_prop(args, session, args.work + "/andromeda.prop")
    if os.path.exists(vendor_andromeda_prop):
        helpers.mount.bind_file(args, args.work + "/andromeda.prop", vendor_andromeda_prop)
    elif os.path.exists(vendor_waydroid_prop):
        helpers.mount.bind_file(args, args.work + "/andromeda.prop", vendor_waydroid_prop)

def umount_rootfs(args):
    helpers.mount.umount_all(args, tools.config.defaults["rootfs"])
